import json
import logging
from typing import Callable

from langchain_core.language_models import BaseChatModel

from ceo.ability.ability import Ability
from ceo.prompt import (
    SchedulerPrompt,
    AnalyserPrompt,
    ExecutorPrompt,
    IntrospectionPrompt,
    QueryResolverPrompt,
    SelfIntroducePrompt
)

log = logging.getLogger('ceo')


class BaseAgent:
    def __init__(self, abilities: list[Callable], brain: BaseChatModel, name: str, query: str = ''):
        self._abilities = list()
        self._act_count = 0
        self._name = name
        self._model = brain
        self._query_high_level = self._query_by_step = str()
        if query is not None and query != '':
            self._query_high_level, self._query_by_step = QueryResolverPrompt(query).invoke(self._model)
        for ability in abilities:
            self._abilities.append(Ability(ability))
        self._introduction = str()
        self.introduce()
        self.__prev_results = list()
        self.__schedule = list()

    @property
    def abilities(self) -> list[Ability]:
        return self._abilities

    @property
    def name(self) -> str:
        return self._name

    @property
    def introduction(self) -> str:
        return self._introduction

    @property
    def brain(self) -> BaseChatModel:
        return self._model

    def __repr__(self):
        ability_str = '['
        for ability in self._abilities:
            ability_str += f'{ability}, '
        ability_str = ability_str[:-2] + ']'
        if ability_str == ']':
            ability_str = '[]'
        return json.dumps({
            self._name: {
                "name": self._name,
                "brain": self._model.dict()['model_name'],
                "abilities": ability_str
            }
        }, ensure_ascii=False)

    def __str__(self):
        return self.__repr__()

    def introduce(self, update: bool = False) -> str:
        if self._introduction == '' or update:
            self._introduction = SelfIntroducePrompt(agent=self).invoke(self._model)
        return self._introduction

    def grant_ability(self, ability: Callable, update_introduction: bool = True):
        self._abilities.append(Ability(ability))
        self.introduce(update_introduction)

    def grant_abilities(self, abilities: list[Callable]):
        for ability in abilities:
            self.grant_ability(ability, update_introduction=False)
        self.introduce(update=True)

    def deprive_ability(self, ability: Callable, update_introduction: bool = True):
        ability = Ability(ability)
        for _ability in self._abilities:
            if _ability.name == ability.name:
                self._abilities.remove(_ability)
        self.introduce(update_introduction)

    def deprive_abilities(self, abilities: list[Callable]):
        for ability in abilities:
            self.deprive_ability(ability, update_introduction=False)
        self.introduce(update=True)

    def plan(self, _log: bool = True) -> list:
        scheduling = SchedulerPrompt(query=self._query_by_step, abilities=self._abilities)
        self.__schedule = scheduling.invoke(self._model)
        if _log:
            log.debug(f'Agent: {self._name}; Schedule: {[_.name for _ in self.__schedule]}; Query: "{self._query_high_level}";')
        return self.__schedule

    def reposition(self):
        self.__prev_results = list()
        self.__schedule = list()
        self._act_count = 0
        return self

    def assign(self, query: str):
        self._query_high_level, self._query_by_step = (
            QueryResolverPrompt(query=query).invoke(self._model))
        return self.reposition()

    def reassign(self, query: str):
        return self.assign(query)

    def __step_quiet(self) -> str:
        if self._act_count < len(self.__schedule):
            analysing = AnalyserPrompt(
                query=self._query_by_step,
                prev_results=self.__prev_results,
                action=self.__schedule[self._act_count]
            )
            action, params = analysing.invoke(self._model)
            executing = ExecutorPrompt(params=params, action=action)
            action_str = f'Agent: {self._name}; Action {self._act_count + 1}/{len(self.__schedule)}: {executing.invoke(model=self._model)};'
            self.__prev_results.append(action_str)
            self._act_count += 1
            log.debug(action_str)
            return action_str
        self.reposition()
        return ''

    def just_do_it(self) -> str | None:
        if not self.plan():
            return None
        for act_count in range(len(self.__schedule)):
            self.__step_quiet()
        response = IntrospectionPrompt(
            query=self._query_high_level,
            prev_results=self.__prev_results,
        ).invoke(self._model)
        log.debug(f'Agent: {self._name}; Conclusion: {response};')
        self.reposition()
        return f'{self._name}: {response}'



import json
import logging
from collections.abc import Iterator

from langchain_core.language_models import BaseChatModel

from ceo.prompt.prompt import Prompt

log = logging.getLogger('ceo.prompt')


class IntrospectionPrompt(Prompt):
    def __init__(self, query: str, prev_results: list | str, ext_context: str = ''):
        prompt = json.dumps({
            "precondition": "Below are actions you (you are the bot/assistant) have performed to achieve the [user_query]. "
                            "You are talking to the user (I (who now talking to you here) am the user), "
                            "use 'you' instead of the 'user' when you organize your response (output).",
            "user_query": f'"{query}"',
            "task": "Tell user's intention first, "
                    "then think seriously whether you have achieved the user's query according to actions you have performed. "
                    "Finally, provide the results wanted by the user based on [user query]. "
                    "If you did not achieve the user's query, explain why?",
            "output_format": "text",
            "output_example": "Your intention is to calculate math, and I was trying to open calculator. "
                              "But I failed because I did not have that ability to open calculator. "
                              "I have not achieved your intention.",
            "hint_for_output": "Your output should be concise and comprehensive",
            "actions_performed": prev_results if isinstance(prev_results, str) else str(prev_results)
        }, ensure_ascii=False)
        super().__init__(prompt, ext_context)
        log.debug(f'IntrospectionPrompt: {self.prompt}')

    def invoke(self, model: BaseChatModel, stream: bool = False) -> str | Iterator:
        if stream:
            return model.stream(self.prompt)
        return model.invoke(self.prompt).content



import json
import logging
from collections.abc import Iterator

from langchain_core.language_models import BaseChatModel

from ceo.prompt.prompt import Prompt

log = logging.getLogger('ceo.prompt')


class SelfIntroducePrompt(Prompt):
    def __init__(self, agent: any, ext_context: str = ''):
        prompt = json.dumps({
            "task": "Introduce yourself briefly based on the information provided. "
                    "Only tell what you exactly can do based on your abilities.",
            "information": agent.__repr__(),
            "output_format": "My name is <name>. What can I do: ..."
        }, ensure_ascii=False)
        super().__init__(prompt, ext_context)
        log.debug(f'SelfIntroducePrompt: {self.prompt}')

    def invoke(self, model: BaseChatModel, stream: bool = False) -> str | Iterator:
        if stream:
            return model.stream(self.prompt)
        return model.invoke(self.prompt).content



import json
import logging

from langchain_core.language_models import BaseChatModel

from ceo.ability.ability import Ability
from ceo.prompt.prompt import Prompt

log = logging.getLogger('ceo.prompt')


class SchedulerPrompt(Prompt):
    def __init__(self, query: str, abilities: list[Ability], ext_context: str = ''):
        self.abilities = abilities
        prompt = dict()
        for ability in self.abilities:
            prompt[ability.name] = str(ability)
        prompt = json.dumps({
            "precondition": "Below are the tools(or abilities) you can use "
                            "(you can only use the following tools(or abilities)). "
                            f'Now there is a [user_query].',
            "user_query": f'"{query}"',
            "task": "What you need to do is to plan your workflow based on the [tools] and [user_query].",
            "description": "[user_query] might contains many steps, "
                           "think carefully about every step and plan your workflow based on your tools.",
            "hint_for_tool_usage": "[user_query] sometimes need to use one specific tool(s) more than once, "
                                   "you need to estimate as accurately as possible "
                                   "the number of times specific tools need to be used "
                                   "to properly achieve the [user_query]!",
            "hint_for_tool_choosing": "Sometimes some of the tools are irrelevant to [user_query]. "
                                      "Make sure to choose tools properly and wisely.",
            "output_format": "{your_thinking_process}\nschedule:{your_schedule_as_a_list_of_tool_names}",
            "hint_for_output_format": 'firstly, output your thinking process step by step clear and organized.'
                                      'secondly, outputs a list(python_list_format) of names of tools, surrounded by "[ ]", split by ", ", '
                                      'you can refer to [output_example].',
            "output_example": "1.First, I need to determine which ingredients to purchase, "
                              "which requires checking a recipe or personal preferences to decide.\n"
                              "2.After determining the ingredients, "
                              "I need to go to the market or supermarket to buy the required ingredients.\n"
                              "3.After purchasing the ingredients, I need to bring them home.\n"
                              "4.Once home, I need to wash and prepare the ingredients.\n"
                              "5.After preparation, I start cooking.\n"
                              "6.After cooking is complete, I need to arrange the dishes on the dining table.\n"
                              "schedule:[go_to_market, payment_purchase, go_home, do_wash, do_cook, arrange_dished_on_table]",
            "tools(abilities)": json.dumps(prompt, ensure_ascii=False)
        }, ensure_ascii=False)
        super().__init__(prompt, ext_context)
        log.debug(f'SchedulerPrompt: {self.prompt}')

    def invoke(self, model: BaseChatModel) -> list[Ability]:
        results = model.invoke(self.prompt).content
        if not results.startswith('['):
            results = results[results.find('['):]
        if not results.endswith(']'):
            results = results[:results.rfind(']') + 1]
        results = results[1:-1].split(',')
        _fin_results = list()
        for _a_result in results:
            for ability in self.abilities:
                if ability.name == _a_result.strip():
                    _fin_results.append(ability)
        return _fin_results



import abc


class MemoryAugment:
    def __init__(self, memory: dict | None = None):
        self._memory = memory

    @property
    def memory(self) -> dict:
        return self._memory

    @abc.abstractmethod
    def bring_in_memory(self, memory: dict):
        pass



import json
import logging
from collections.abc import Iterator

from langchain_core.language_models import BaseChatModel

from ceo.ability.ability import Ability
from ceo.prompt.prompt import Prompt

log = logging.getLogger('ceo.prompt')


class ExecutorPrompt(Prompt):
    def __init__(self, params: dict, action: Ability, ext_context: str = ''):
        self.action = action
        self.params = params
        prompt = json.dumps({
            "precondition": "Below is a tool and your choice (params) for using this tool.",
            "task": "Explain what you are going to do.",
            "output_format": "text",
            "output_example": "I am trying to open calculator.",
            "tool": action.__repr__(),
            "params (choice)": str(params)
        }, ensure_ascii=False)
        super().__init__(prompt, ext_context)
        log.debug(f'ExecutorPrompt (before): {self.prompt}')

    def explain(self, model: BaseChatModel, stream: bool = False) -> str | Iterator:
        if stream:
            return model.stream(self.prompt)
        return model.invoke(self.prompt).content

    def invoke(self, model: BaseChatModel, stream: bool = False) -> str | Iterator:
        result = self.action.__call__(**self.params)
        prompt = json.dumps({
            "precondition": "Below is a tool, your choice (params) for the tool, "
                            "and the result of your using of this tool.",
            "task": "Explain what you have done, and write down the result detailed. "
                    "The result is shown below at [result].",
            "output_format": "text",
            "output_contains": [
                "{the_tool(ability)_you_used}",
                "{the_choice_you_made}",
                "{what_you_have_done}"
            ],
            "hint_for_output": 'When you give the response, say "ability" instead of "tool".',
            "output_example": "I wrote a wechat message which says 'Bonjour'. The result is 'success'.",
            "tool": self.action.__repr__(),
            "params (choice)": str(self.params),
            "result": str(result)
        }, ensure_ascii=False)
        prompt = f'{self.ext_context}{self.seperator}{prompt}'
        log.debug(f'ExecutorPrompt (after): {prompt}')
        if stream:
            return model.stream(prompt)
        return model.invoke(prompt).content



import json
import logging

from langchain_core.language_models import BaseChatModel

from ceo.ability import Ability
from ceo.prompt.prompt import Prompt

log = logging.getLogger('ceo.prompt')

SEPARATOR = "----SEPARATOR----"

OUTPUT_EXAMPLE = """
Step 1: In the provided history, the events related to the user's query are as follows, listed chronologically:
        1. Buying two tomatoes: This event is the first in the sequence and is directly related to the user's query as it involves the acquisition of the main ingredient needed for the subsequent steps.
        2. Going home: Following the purchase, the user returns home, which is a necessary step before proceeding with the cooking process.
        
        From the history, we extract the following information related to the user's query:
        1. The user has successfully completed the purchase of tomatoes.
        2. The user is now at home, which is the location where the next steps of the process will take place.
        
        Based on these details, the subsequent steps to complete are:
        1. Cook the tomatoes using a frying pan.
        2. Place the cooked tomatoes on the dining table.

Step 2: The user query is "Help me buy two tomatoes, then after getting home, cook the tomatoes using a frying pan, 
        and finally place the cooked tomatoes on the dining table." According to the history, the first two parts have been completed, 
        but the last two parts (cooking and placing) are still pending. Therefore, the user query has not been fully and properly accomplished.

Step 3: The unfinished parts of the user query are "cook the tomatoes using a frying pan" and "place the cooked tomatoes on the dining table." 
        The abilities I possess include "go_home()", "do_cook(ingredient: str, cooking_utensils: str) -> bool", and "arrange_dished()". 
        Among these, "do_cook" can complete the cooking task, 
        while "arrange_dished" can complete the placing task. Thus, my abilities can fulfill the unfinished parts of the user query.

Step 4: Since the user's query is not fully accomplished, and I have the ability to cook and arrange dishes, 
        my next move is to cook the tomatoes. I will use the "do_cook" ability with "tomatoes" as the ingredient and "frying pan" as the cooking utensil 
        because these parameters align with the user's request to cook the tomatoes using a frying pan.

Step 5: This step is not applicable because the user query has not been fully accomplished, 
        and I have the ability to continue progressing.

Step 6: This step is not applicable because the user query has not been fully accomplished.

""" + SEPARATOR + """

params:{
  "ingredient": "tomatoes",
  "cooking_utensils": "frying pan"
}

ability:[do_cook]
"""


class NextMovePrompt(Prompt):
    def __init__(self, query: str,
                 abilities: list[Ability], history: str = '',
                 ext_context: str = ''):
        self.abilities = abilities
        abilities_dict = dict()
        for ability in self.abilities:
            abilities_dict[ability.name] = str(ability)
        if history == '':
            history = "Nothing happened before you."
        prompt = json.dumps({
            "precondition": "Below are the abilities you have(you can only use the following abilities)."
                            "[history] shows events happened before you. And there is a [user_query].",
            "user_query": f'"{query}"',
            "abilities": json.dumps(abilities_dict, ensure_ascii=False),
            "history": history,
            "instructions_you_must_follow_step_by_step": [{
                    "step": 1,
                    "first_action": "List all events in the [history] related to [user_query] respectively and chronologically.",
                    "second_action": "Extract and list all key information related to [user_query] from [history] "
                                     "formatted one by one respectively.",
                    "additional": "For any details mentioned in [history] about [user_query], you should preserve them in full, "
                                  "especially specific information with accuracy requirements such as numbers, dates, etc."
                }, {
                    "step": 2,
                    "action": "Analyse whether the [user_query] has been fully and properly accomplished "
                              "according to your report in [step_1], "
                              "and provide your analysis process and basis."
                }, {
                    "step": 3,
                    "condition": "If the [user_query] has not been fully properly accomplished",
                    "action": "Analyse whether your [abilities] can complete the unfinished part of the [user_query], "
                              "and provide the basis according to your report in [step_1]."
                }, {
                    "step": 4,
                    "condition": "If the [user_query] has not been fully properly accomplished and "
                                 "there is an ability in your [abilities] "
                                 "that can further advance the accomplishment of the [user_query] based on [history].",
                    "first_action": "Plan and explain your next move based on [history] "
                                    "for further advancing the [user_query].",
                    "second_action": "Choose and provide the ability according to your next move"
                                     "(only one single ability can be chosen)",
                    "third_action": "After you have chosen the ability as next move, "
                                    "generate values of parameters for the ability(function) to achieve [next move], "
                                    "before you generate values of parameters, explain why you give these values to params.",
                }, {
                    "step": 5,
                    "condition": "If the [user_query] has not been fully properly accomplished and "
                                 "there is no ability in your [abilities] "
                                 "that can further advance the accomplishment of the [user_query]",
                    "action": 'Provide a special ability called "-mission-failed-" (which is not a real ability).'
                }, {
                    "step": 6,
                    "condition": "If the [user_query] has been fully and properly accomplished "
                                 "according to [history] and [user_query]",
                    "action": 'Provide a special ability called "-mission-complete-" (which is not a real ability).'
                }
            ],
            "output_format": "{step1_thought_process}\n{step2_thought_process}\n"
                             "{step3_thought_process}\n{step4_thought_process}\n"
                             "{step5_thought_process}\n{step6_thought_process}\n"
                             f"{SEPARATOR}\n"
                             'params:{'
                             '{name_of_param_1}:{value_for_param_1},'
                             '{name_of_param_2}:{value_for_param_2},'
                             '{name_of_param_...}:{value_for_param_...}'
                             '}\n'
                             "ability:[ability.name]",
            "hint_for_thought_process_output": "Thought processes of all steps(from 1 to 6) should be output.",
            "hint_for_ability_choosing": "Only one single ability can be chosen.",
            "hint_for_params_output_format": f'The "{SEPARATOR}" pattern should be after '
                                             f'all the thought processes and before the [params and ability].'
                                             f'The params should be after the "{SEPARATOR}" pattern.'
                                             'The params should be formatted as json.'
                                             'The params only gives the params for the chosen one ability.'
                                             'The ability should be after params.',
            "hint_for_separation_pattern": f'The "{SEPARATOR}" pattern which separates [thought processes] and '
                                           '[params, ability] is absolutely important, do not forget to place it.',
            "hint_for_ability_output_format": 'The ability should be after the params. '
                                              'The ability name should be surrounded by "[ ]".',
            "output_example": OUTPUT_EXAMPLE
        }, ensure_ascii=False)
        super().__init__(prompt, ext_context)
        log.debug(f'NextMovePrompt: {self.prompt}')

    def invoke(self, model: BaseChatModel, stream: bool = False) -> tuple[Ability, dict] | bool:
        result: str = model.invoke(self.prompt).content
        log.debug(f"Next move thought process: \n{result}")
        result = result[result.rfind(SEPARATOR):]
        ability_name: str = result[result.find('['):result.rfind(']') + 1].strip()[1:-1]
        params: dict = json.loads(result[result.find('{'):result.rfind('}') + 1].strip())
        if ability_name.__contains__('-mission-complete-'):
            return True
        for ability in self.abilities:
            if ability.name == ability_name:
                return ability, params
        return False



import json
import logging
import random
import datetime
from typing import Callable, override

from langchain_core.language_models import BaseChatModel

from ceo.ability.agentic_ability import PREFIX as AGENTIC_ABILITY_PREFIX
from ceo.brain.base_agent import BaseAgent
from ceo.brain.memory_augment import MemoryAugment
from ceo.enum.Personality import Personality
from ceo.prompt import (
    NextMovePrompt,
    ExecutorPrompt,
    IntrospectionPrompt
)

PRUDENT_P = 0.25
PRUDENT_BETA = 1.45
INQUISITIVE_P = 0.05
INQUISITIVE_BETA = 1.25
log = logging.getLogger('ceo')


class Agent(BaseAgent, MemoryAugment):
    def __init__(self, abilities: list[Callable],
                 brain: BaseChatModel, name: str,
                 personality: Personality = Personality.PRUDENT,
                 query: str = '', memory: dict | None = None):
        BaseAgent.__init__(self, abilities=abilities, brain=brain, name=name, query=query)
        MemoryAugment.__init__(self, memory=memory)
        self.__expected_step = 0
        if personality == Personality.PRUDENT:
            self._p = self.__base_p = PRUDENT_P
            self._beta = PRUDENT_BETA
        elif personality == Personality.INQUISITIVE:
            self._p = self.__base_p = INQUISITIVE_P
            self._beta = INQUISITIVE_BETA

    @property
    def p(self) -> float:
        return self._p

    @property
    def base_p(self) -> float:
        return self.__base_p

    @property
    def beta(self) -> float:
        return self._beta

    @override
    def bring_in_memory(self, memory: dict):
        self._memory.update(memory)
        return self

    @override
    def reposition(self):
        BaseAgent.reposition(self)
        self._memory = dict()
        self.__expected_step = 0
        self._p = self.__base_p
        return self

    @override
    def assign(self, query: str):
        BaseAgent.assign(self, query)
        return self.reposition()

    @override
    def reassign(self, query: str):
        return self.assign(query)

    @override
    def just_do_it(self) -> dict:
        self.estimate_step()
        stop = False
        while True:
            if self._act_count > self.__expected_step:
                stop = self.stop()
                self.penalize()
            _history = json.dumps(self._memory, ensure_ascii=False)
            next_move = False
            if not stop:
                next_move = NextMovePrompt(
                    query=self._query_by_step,
                    abilities=self._abilities,
                    history=_history
                ).invoke(self._model)
                if not isinstance(next_move, bool):
                    action, params = next_move
                    if action.name.startswith(AGENTIC_ABILITY_PREFIX):
                        params = {'query': self._query_by_step, 'memory': self._memory}
                    self.memorize(ExecutorPrompt(params=params, action=action).invoke(model=self._model))
                    self._act_count += 1
                    continue
            response = IntrospectionPrompt(
                query=self._query_high_level,
                prev_results=_history,
            ).invoke(self._model)
            self.reposition()
            log.debug(f'Agent: {self._name}; Conclusion: {response};')
            return {
                "success": next_move,
                "response": response
            }

    def assign_with_memory(self, query: str, memory: dict):
        return self.assign(query).bring_in_memory(memory)

    def estimate_step(self):
        if self._query_by_step == '':
            self.__expected_step = 0
            return
        self.__expected_step = len(self.plan(_log=False))
        log.debug(f'Agent: {self._name}; Expected steps: {self.__expected_step}; Query: "{self._query_high_level}";')

    def memorize(self, action_performed: str):
        now = datetime.datetime.now().strftime('%m/%d/%Y %H:%M:%S.%f')
        new_memory = {
            "date_time": now,
            "agent_name": self._name,
            f"message_from_{self._name}": action_performed
        }
        self._memory[f"{self._name} at {now}"] = new_memory
        log.debug(f'Agent: {self._name}; Memory update: {new_memory};')

    def stop(self) -> bool:
        resample = 3
        log.debug(f'Agent: {self._name}; Termination Probability(p): {self._p}; Penalty Rate(beta): {self._beta};')
        rand_sum = 0.0
        for i in range(resample):
            rand_sum += random.uniform(0, 1)
        rand_avg = rand_sum / resample
        if rand_avg > self._p:
            return False
        return True

    def penalize(self):
        self._p = (self._beta * self._p) % 1.0

    def set_penalty(self, p: float, beta: float):
        self._p = self.__base_p = p
        self._beta = beta
        return self

    def change_personality(self, personality: Personality):
        if personality == Personality.PRUDENT:
            return self.set_penalty(p=PRUDENT_P, beta=PRUDENT_BETA)
        elif personality == Personality.INQUISITIVE:
            return self.set_penalty(p=INQUISITIVE_P, beta=INQUISITIVE_BETA)



import json
import logging

from langchain_core.language_models import BaseChatModel

from ceo.prompt.prompt import Prompt

log = logging.getLogger('ceo.prompt')


class QueryResolverPrompt(Prompt):
    def __init__(self, query: str, ext_context: str = ''):
        prompt = json.dumps({
            "precondition": f'There is a user query: "{query}"',
            "task": "What you need to do is to tell user's intention based on [user query].",
            "task_redeclare": "To tell user's intention based on [user query]. Not your (you are the assistant) intention.",
            "additional": "For any details mentioned by the user, you should preserve them in full, "
                          "especially specific information with accuracy requirements such as numbers, dates, etc.",
            "hint_for_thinking": "Deduce and analyse the [user query] step by step. "
                                 "Keep track of the steps' interdependence and orderliness.",
            "output_format (json)": {
                "step_{n}": "{action_of_step_{n}]"
            },
            "hint_for_output": "Break user's intention(s) down into multiple minimum steps as granular as possible. "
                               "Keep track of the steps' interdependence and orderliness again.",
            "output_example": {
                "step_1": "Open the door",
                "step_2": "(Door opened) Go into the room",
                "step_3": "(Walked in the room) Find the toys in the room",
                "step_...": "(Found toys)..."
            }
        }, ensure_ascii=False)
        self.__query = query
        super().__init__(prompt, ext_context)
        log.debug(f'QueryResolverPrompt: {self.prompt}')

    def invoke(self, model: BaseChatModel) -> tuple[str, str]:
        if self.__query == '':
            return (json.dumps({"User's intention": "Don't do anything."}),
                    json.dumps({"User's query (Step by step)": "Don't do anything."}))
        user_query_by_step = model.invoke(self.prompt).content
        summary_prompt = json.dumps({
            "task": "Summarize user's query into a short sentence "
                    "which includes all the key information from user's query "
                    "(User's query is provided below at [user's query]).",
            "user_query": f'"{user_query_by_step}".',
            "output_format": "string(summarization of [user's query])",
            "output_example": "To find toys for you in the room."
        }, ensure_ascii=False)
        summary = model.invoke(summary_prompt).content
        return (json.dumps({"User's intention": summary}),
                json.dumps({"User's query (Step by step)": user_query_by_step}))



import inspect
import json
from typing import Callable


class Ability:
    def __init__(self, function: Callable):
        signature = inspect.signature(function)
        self.name: str = function.__name__
        self.description: str = inspect.getdoc(function)
        self.function: Callable = function
        self.parameters: dict = dict()
        self.returns: str = str(signature.return_annotation)
        for name, param in signature.parameters.items():
            self.parameters[name] = str(param.annotation)

    def __repr__(self):
        return json.dumps({
            'name': self.name,
            'description': self.description,
            'parameters': self.parameters,
            'returns': self.returns
        }, ensure_ascii=False)

    def __str__(self):
        return self.__repr__()

    def __call__(self, *args, **kwargs):
        return self.function(*args, **kwargs)



import logging
from typing import Callable

from langchain_core.language_models import BaseChatModel

from ceo import get_openai_model
from ceo.prompt import DocstringPrompt

log = logging.getLogger('ceo.ability')


def docstring_generator(func: Callable, brain: BaseChatModel):
    if func.__doc__ in ('', None):
        func.__doc__ = DocstringPrompt(func).invoke(brain)
        log.debug(f'Docstring generated for {func.__name__}. Docstring: "{func.__doc__}"')
    return func


def ability(brain: BaseChatModel):
    if callable(brain) and not isinstance(brain, BaseChatModel):
        def decorator(func):
            return docstring_generator(func, get_openai_model())
        return decorator(brain)

    def decorator(func):
        return docstring_generator(func, brain)
    return decorator



import json
import logging
from typing import override

from ceo.brain.base_agent import BaseAgent
from ceo.ability import Ability
from ceo.brain.memory_augment import MemoryAugment

log = logging.getLogger('ceo.ability')
PREFIX = '__AgenticAbility__'


class AgenticAbility(Ability):
    def __init__(self, agent: BaseAgent | MemoryAugment):
        try:
            _test_mem = agent.memory
        except AttributeError:
            raise TypeError("The 'agent' of AgenticAbility should be instance of 'ceo.Agent'.")
        self._agent = agent
        self.__name__ = f'{PREFIX}talk_to_{agent.name}'
        self.__doc__ = json.dumps({
            "description": {
                "brief_description": f'Initiates a conversation with "{agent.name}" to use its abilities.',
                "detailed_description": f"First, carefully consider and explore {agent.name}'s potential abilities in solving your tasks, "
                                        f"then, if you need {agent.name}'s help, you must tell comprehensively, precisely "
                                        f"and exactly what you need {agent.name} to do.",
                f"self_introduction_from_{agent.name}": agent.introduction,
                "hint": f"By reading [self_introduction_from_{agent.name}], you can learn what {agent.name} can do, "
                        f"and then decide whether to initiates a conversation with {agent.name} according to its abilities.",
                "args": [{
                    "query": {
                        "name": "query",
                        "type": "str",
                        "description": f"A comprehensively, precisely and exactly instruction to be processed by {agent.name}."
                    }
                  }
                ],
                "returns": {
                  "type": "str",
                  "description": f"{agent.name}'s response to your instruction."
                }
            }
        }, ensure_ascii=False)
        super().__init__(self)
        log.debug(f'Agent dispatcher generated. {self.__name__}: {self.__doc__}')

    @override
    def __call__(self, query: str, memory: dict, *args, **kwargs) -> str:
        self._agent.assign(query)
        self._agent.bring_in_memory(memory)
        result = self._agent.just_do_it()
        if isinstance(result, dict):
            return json.dumps(result, ensure_ascii=False)
        return result
